-----------------------------------------------------------------------
-- To do and/or ideas
--
-- From Rabbit:
-- * Parse damage messages from swstats and damagemeters.
-- * Automatically output our damage done and position to chat when a boss has
-- been killed; this is easy with Babble-Boss, CHAT_MSG_HOSTILE_DEATH and
-- UnitClassification.
-- * Maybe add an auto-reset for boss kill as well.
--
-- From Nymbia:
-- * Pick up the slack of users without the addon by splitting up the members
-- without the addon among those with it.  ie, in a 25 man raid where 10 users
-- have Violation, each of them is assigned to parse the damage of either 1 or
-- 2 other raid members.
-- * Possibly keep two running totals:  'this fight' and 'overall', to allow
-- for statistics to be displayed for a single encounter without requiring
-- constant resets.
--
-- From <person>:
-- * Brew my coffee when I win the damage meter contest after a boss fight.
--
-----------------------------------------------------------------------

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = AceLibrary("AceLocale-2.2"):new("Violation")

L:RegisterTranslations("enUS", function() return {
	["Resetting data as requested by %s."] = true,

	["Track damage"] = true,
	["Whether or not to track damage done."] = true,
	["Track healing"] = true,
	["Whether or not to track healing done."] = true,
	["Reset"] = true,
	["Options for when to reset the data."] = true,
	["Reset now"] = true,
	["Immediately resets your data."] = true,
	["Reset group"] = true,
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = true,
	["Leader resets"] = true,
	["Accept reset requests from raid leaders or officers."] = true,
	["Auto-reset"] = true,
	["Entering combat"] = true,
	["Automatically reset the data when entering combat."] = true,
	["Leaving combat"] = true,
	["Automatically reset the data when leaving combat."] = true,

	["Text"] = true,
	["What data to show on the FuBar panel, if available."] = true,
	["Show damage done on the FuBar panel."] = true,
	["Show healing done on the FuBar panel."] = true,
	["Print on reset"] = true,
	["Prints your current data locally when your data is reset."] = true,

	["Damage done: %d (#%s)."] = true,
	["Healing done: %d (#%s)."] = true,
	["D: "] = true,
	["H: "] = true,

	["Healing"] = true,
	["Damage"] = true,
	["Player"] = true,
	["#"] = true,
	["No data gathered yet."] = true,
	["|cffeda55fClick|r to reset."] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Resetting data as requested by %s."] = "%s의 요청으로 데이터를 초기화 합니다.",
	["Reset"] = "초기화",
	["Options for when to reset the data."] = "데이터 초기화를 위한 설정",
	["Reset now"] = "바로 초기화",
	["Immediately resets your data."] = "데이터를 바로 초기화 합니다.",
	["Reset group"] = "파티 초기화",
	["Resets the Violation data for every group member. Only the group leader or an officer can do this."] = "파티원이 갱신될 때마다 초기화합니다. 이것은 파티장이나 공격대장만이 가능합니다.",
	["Leader resets"] = "리더 초기화",
	["Accept reset requests from raid leaders or officers."] = "공격대장의 요청이 있을 때 데이터를 초기화를 수락합니다.",
	["Auto-reset"] = "자동 초기화",
	["Entering combat"] = "전투에 들어갈 때",
	["Automatically reset the data when entering combat."] = "전투가 시작될 때 자동으로 초기화 합니다.",
	["Leaving combat"] = "전투가 끝날 때",
	["Automatically reset the data when leaving combat."] = "전투가 끝날 때 자동으로 초기화 합니다",

	["Show data"] = "데이터 보기",
	["Shows your own data on the FuBar panel, if available."] = "가능하면, 푸바 패널에 자신의 피해량만 표시합니다.",
	["Print damage on reset"] = "초기화시 피해량 표시",
	["Prints your current damage locally when your data is reset."] = "모든 피해량을 초기화할 때 기록을 위해 채팅창에 피해량을 표시합니다.",

	["Resetting data. Damage done: %d (#%s)."] = "데이터를 초기화 합니다. 총 피해입힌 양: %d (#%s).",
	["No data gathered yet."] = "아직 수집한 데이터가 없습니다.",
	["|cffeda55fClick|r to reset."] = "|cffeda55f클릭|r하면 초기화 합니다.",
} end)

-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------
local tablet = AceLibrary("Tablet-2.0")
local BC = AceLibrary("Babble-Class-2.2")
local roster = AceLibrary("Roster-2.1")

local playerName = nil

local coloredNames = setmetatable({}, {__index =
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return nil end
		self[key] = "|cff" .. BC:GetHexColor(obj.class) .. key .. "|r"
		return self[key]
	end
})

local damage = nil
local healing = nil

local damageDirty = 0
local healingDirty = 0

local damagePosition = nil
local healingPosition = nil

local db = nil

local _G = getfenv(0)

local red = "Interface\\AddOns\\Violation\\icons\\red"
local blue = "Interface\\AddOns\\Violation\\icons\\blue"
local grey = "Interface\\AddOns\\Violation\\icons\\grey"

local percentString = "(%d%%)"

-----------------------------------------------------------------------
-- Local heap
-----------------------------------------------------------------------

local new, del
do
	local cache = setmetatable({},{__mode='k'})
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select("#", ...) do
				table.insert(t, (select(i, ...)))
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == "table" then
			for k, v in pairs(t) do
				if type(v) == "table" then
					del(v)
				end
				t[k] = nil
			end
			cache[t] = true
		end
		return nil
	end
end

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local options = {
	type = "group",
	args = {
		damage = {
			type = "toggle",
			name = L["Track damage"],
			desc = L["Whether or not to track damage done."],
			get = function() return Violation.db.profile.trackDamage end,
			set = function(v)
				Violation:ToggleTrackingDamage(v)
				Violation:UpdateDisplay()
			end,
			order = 100,
		},
		healing = {
			type = "toggle",
			name = L["Track healing"],
			desc = L["Whether or not to track healing done."],
			get = function() return Violation.db.profile.trackHealing end,
			set = function(v)
				Violation:ToggleTrackingHealing(v)
				Violation:UpdateDisplay()
			end,
			order = 101,
		},
		text = {
			type = "group",
			name = L["Text"],
			desc = L["What data to show on the FuBar panel, if available."],
			disabled = function() return not FuBar end,
			args = {
				damage = {
					type = "toggle",
					name = L["Damage"],
					desc = L["Show damage done on the FuBar panel."],
					get = function() return Violation.db.profile.showDamageText end,
					set = function(v)
						Violation.db.profile.showDamageText = v
						Violation:UpdateText()
					end,
					order = 1,
				},
				healing = {
					type = "toggle",
					name = L["Healing"],
					desc = L["Show healing done on the FuBar panel."],
					get = function() return Violation.db.profile.showHealingText end,
					set = function(v)
						Violation.db.profile.showHealingText = v
						Violation:UpdateText()
					end,
					order = 2,
				},
			},
			order = 102,
		},
		spacer = {
			type = "header",
			name = " ",
			order = 150,
		},
		print = {
			type = "toggle",
			name = L["Print on reset"],
			desc = L["Prints your current data locally when your data is reset."],
			get = function() return Violation.db.profile.printOnReset end,
			set = function(v) Violation.db.profile.printOnReset = v end,
			order = 200,
		},
		reset = {
			type = "group",
			name = L["Reset"],
			desc = L["Options for when to reset the data."],
			args = {
				now = {
					type = "execute",
					name = L["Reset now"],
					desc = L["Immediately resets your data."],
					func = function() Violation:Reset() end,
					order = 100,
				},
				group = {
					type = "execute",
					name = L["Reset group"],
					desc = L["Resets the Violation data for every group member. Only the group leader or an officer can do this."],
					func = function() Violation:ResetGroup() end,
					disabled = function()
						return not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader()
					end,
					order = 101,
				},
				leader = {
					type = "toggle",
					name = L["Leader resets"],
					desc = L["Accept reset requests from raid leaders or officers."],
					get = function() return Violation.db.profile.leaderReset end,
					set = function(v) Violation.db.profile.leaderReset = v end,
					order = 102,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 150,
				},
				autoheader = {
					type = "header",
					name = L["Auto-reset"],
					order = 200,
				},
				gaincombat = {
					type = "toggle",
					name = L["Entering combat"],
					desc = L["Automatically reset the data when entering combat."],
					get = function() return Violation.db.profile.gainCombat end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_DISABLED")
						end
						Violation.db.profile.gainCombat = v
					end,
					order = 201,
				},
				leavecombat = {
					type = "toggle",
					name = L["Leaving combat"],
					desc = L["Automatically reset the data when leaving combat."],
					get = function() return Violation.db.profile.leaveCombat end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_ENABLED")
						end
						Violation.db.profile.leaveCombat = v
					end,
					order = 202,
				},
			},
			order = 201,
		},
	},
}

-----------------------------------------------------------------------
-- Addon initialization
-----------------------------------------------------------------------
Violation = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceComm-2.0", "FuBarPlugin-2.0", "Parser-3.0")
local Violation = Violation

Violation.hasIcon = true

function Violation:OnInitialize()
	local revision = tonumber(("$Revision: 28472 $"):sub(12, -3)) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.VIOLATION_VERSION = self.version
	_G.VIOLATION_REVISION = self.revision

	self:RegisterDB("ViolationDB")
	self:RegisterDefaults("profile", {
		gainCombat = nil,
		leaveCombat = nil,
		leaderReset = true,
		showDamageText = nil,
		showHealingText = nil,
		printOnReset = true,
		trackDamage = true,
		trackHealing = true,
	})

	self.OnMenuRequest = options
	self.hasNoColor = true
	self.independentProfile = true

	self:RegisterChatCommand({ "/violation" }, self.OnMenuRequest, "VIOLATION")

	self:SetCommPrefix("Violation")
	self:SetDefaultCommPriority("BULK")

	playerName = UnitName("player")

	self:SetIcon(grey)
end

function Violation:OnEnable()
	db = self.db.profile

	self:ToggleTrackingDamage(db.trackDamage)
	self:ToggleTrackingHealing(db.trackHealing)

	self:Reset()

	self:RegisterComm("Violation", "GROUP", "OnComm")

	self:RegisterEvent("CHAT_MSG_ADDON", "Listen")

	if db.gainCombat then
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
	end

	if db.leaveCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
	end

	self:ScheduleRepeatingEvent("Broadcast", self.Broadcast, 5, self)

	self:SetIcon(red)
end

function Violation:OnProfileEnable()
	db = self.db.profile

	self:ToggleTrackingDamage(db.trackDamage)
	self:ToggleTrackingHealing(db.trackHealing)

	if db.gainCombat and not self:IsEventRegistered("PLAYER_REGEN_DISABLED") then
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
	elseif not db.gainCombat then
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	end

	if db.leaveCombat and not self:IsEventRegistered("PLAYER_REGEN_ENABLED") then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
	elseif not db.leaveCombat then
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	end

	self:UpdateDisplay()
end

function Violation:OnDisable()
	self:SetIcon(grey)
end

function Violation:Reset()
	if type(damage) == "table" then
		if db.printOnReset and damage[playerName] > 0 then
			self:Print(string.format(L["Damage done: %d (#%s)."], damage[playerName], damagePosition or "?"))
		end
		damage = del(damage)
	end
	if type(healing) == "table" then
		if db.printOnReset and healing[playerName] > 0 then
			self:Print(string.format(L["Healing done: %d (#%s)."], healing[playerName], healingPosition or "?"))
		end
		healing = del(healing)
	end

	damage = new()
	damage[playerName] = 0

	healing = new()
	healing[playerName] = 0

	self:UpdateDisplay()
end

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local dmgEvent = {
	eventType = "Damage",
	sourceID = "player",
}

local healEvent = {
	eventType = "Heal",
	sourceID = "player",
}

function Violation:ToggleTrackingDamage(input)
	if input and not self:IsParserEventRegistered(dmgEvent) then
		self:RegisterParserEvent(dmgEvent, "OnDamageEvent")
	elseif not input then
		self:UnregisterParserEvent(dmgEvent)
	end
	db.trackDamage = input
end

function Violation:ToggleTrackingHealing(input)
	if input and not self:IsParserEventRegistered(healEvent) then
		self:RegisterParserEvent(healEvent, "OnHealEvent")
	elseif not input then
		self:UnregisterParserEvent(healEvent)
	end
	db.trackHealing = input
end

-----------------------------------------------------------------------
-- UI
-----------------------------------------------------------------------

function Violation:OnClick()
	self:Reset()
end

local function Sorter(a, b)
	return a[2] > b[2]
end

local function AddHealingData()
	local tbl = new()
	for k, v in pairs(healing) do
		if type(v) == "number" and v > 0 then
			table.insert(tbl, new(k, v))
		end
	end

	if #tbl > 0 then
		local cat = tablet:AddCategory(
			"columns", #tbl > 1 and 4 or 3,
			"text", L["#"],
			"text2", L["Player"],
			"text3", L["Healing"],
			"child_justify", "LEFT",
			"child_justify2", "LEFT",
			"child_justify3", "RIGHT",
			"child_justify4", "RIGHT",
			"child_textR", 1,
			"child_textG", 1,
			"child_textB", 0,
			"child_text3R", 1,
			"child_text3G", 1,
			"child_text3B", 1,
			"child_text4R", 0.65,
			"child_text4G", 0.65,
			"child_text4B", 0.65
		)

		table.sort(tbl, Sorter)

		local percentage = nil
		if #tbl > 1 then
			local total = 0
			for i, v in ipairs(tbl) do
				total = total + v[2]
			end
			percentage = new()
			for i, v in ipairs(tbl) do
				table.insert(percentage, math.floor((v[2] * 100) / total))
			end
		end

		local playerShown = nil
		for i, v in ipairs(tbl) do
			if i > 10 then break end -- Only show the top 10.
			if percentage then
				cat:AddLine(
					"text", tostring(i) .. ".",
					"text2", coloredNames[v[1]],
					"text3", tostring(v[2]),
					"text4", string.format(percentString, percentage[i])
				)
			else
				cat:AddLine(
					"text", tostring(i) .. ".",
					"text2", coloredNames[v[1]],
					"text3", tostring(v[2])
				)
			end
			if v[1] == playerName then
				playerShown = true
				if #tbl > 1 then
					healingPosition = i
				else
					healingPosition = nil
				end
			end
		end

		percentage = del(percentage)

		if not playerShown and healing[playerName] > 0 then
			cat:AddLine("text", "?", "text2", coloredNames[playerName], "text3", tostring(healing[playerName]))
			healingPosition = "?"
		end
	else
		local cat = tablet:AddCategory(
			"columns", 1,
			"text", L["Healing"]
		)
		cat:AddLine("text", L["No data gathered yet."])
	end

	-- Cleanup.
	tbl = del(tbl)
end

local function AddDamageData()
	local tbl = new()
	for k, v in pairs(damage) do
		if type(v) == "number" and v > 0 then
			table.insert(tbl, new(k, v))
		end
	end

	if #tbl > 0 then
		local cat = tablet:AddCategory(
			"columns", #tbl > 1 and 4 or 3,
			"text", L["#"],
			"text2", L["Player"],
			"text3", L["Damage"],
			"child_justify", "LEFT",
			"child_justify2", "LEFT",
			"child_justify3", "RIGHT",
			"child_justify4", "RIGHT",
			"child_textR", 1,
			"child_textG", 1,
			"child_textB", 0,
			"child_text3R", 1,
			"child_text3G", 1,
			"child_text3B", 1,
			"child_text4R", 0.65,
			"child_text4G", 0.65,
			"child_text4B", 0.65
		)

		table.sort(tbl, Sorter)

		local percentage = nil
		if #tbl > 1 then
			local total = 0
			for i, v in ipairs(tbl) do
				total = total + v[2]
			end
			percentage = new()
			for i, v in ipairs(tbl) do
				table.insert(percentage, math.floor((v[2] * 100) / total))
			end
		end

		local playerShown = nil
		for i, v in ipairs(tbl) do
			if i > 10 then break end -- Only show the top 10.
			if percentage then
				cat:AddLine(
					"text", tostring(i) .. ".",
					"text2", coloredNames[v[1]],
					"text3", tostring(v[2]),
					"text4", string.format(percentString, percentage[i])
				)
			else
				cat:AddLine(
					"text", tostring(i) .. ".",
					"text2", coloredNames[v[1]],
					"text3", tostring(v[2])
				)
			end
			if v[1] == playerName then
				playerShown = true
				if #tbl > 1 then
					damagePosition = i
				else
					damagePosition = nil
				end
			end
		end

		percentage = del(percentage)

		if not playerShown and damage[playerName] > 0 then
			cat:AddLine("text", "?", "text2", coloredNames[playerName], "text3", tostring(damage[playerName]))
			damagePosition = "?"
		end
	else
		local cat = tablet:AddCategory(
			"columns", 1,
			"text", L["Damage"]
		)
		cat:AddLine("text", L["No data gathered yet."])
	end

	-- Cleanup.
	tbl = del(tbl)
end

function Violation:OnTooltipUpdate()
	if db.trackDamage then
		AddDamageData()
	end
	if db.trackHealing then
		AddHealingData()
	end

	tablet:SetHint(L["|cffeda55fClick|r to reset."])
end

function Violation:OnTextUpdate()
	if db and (db.showDamageText or db.showHealingText) then
		local text = new()
		if db.showDamageText and damage[playerName] > 0 then
			table.insert(text, L["D: "] .. (type(damagePosition) ~= "nil" and tostring(damagePosition) .. ". " or "") .. damage[playerName])
		end
		if db.showHealingText and healing[playerName] > 0 then
			table.insert(text, L["H: "] .. (type(healingPosition) ~= "nil" and tostring(healingPosition) .. ". " or "") .. healing[playerName])
		end
		if #text > 0 then
			self:SetText(table.concat(text, "/"))
		else
			self:SetText(self.name)
		end
	else
		self:SetText(self.name)
	end
end

-----------------------------------------------------------------------
-- Data broadcasting and recieving
-----------------------------------------------------------------------
function Violation:Broadcast()
	if damageDirty > 0 then
		self:SendCommMessage("GROUP", "Damage", damageDirty)
		damageDirty = 0
	end
	if healingDirty > 0 then
		self:SendCommMessage("GROUP", "Healing", healingDirty)
		healingDirty = 0
	end
end

function Violation:ResetGroup()
	if not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader() then
		error("Only the group leader or an officer can do that.")
		return
	end

	self:SendCommMessage("GROUP", "LeaderReset")
end

function Violation:OnComm(prefix, sender, distribution, command, data)
	if distribution ~= "GROUP" or type(command) ~= "string" then return end

	if command == "Damage" then
		damage[sender] = (damage[sender] or 0) + tonumber(data)
		self:ScheduleEvent("DisplayUpdate", self.UpdateDisplay, 5, self)
	elseif command == "Healing" then
		healing[sender] = (healing[sender] or 0) + tonumber(data)
		self:ScheduleEvent("DisplayUpdate", self.UpdateDisplay, 5, self)
	elseif command == "LeaderReset" then
		local obj = roster:GetUnitObjectFromName(sender)
		if obj.rank > 0 then
			self:Print(string.format(L["Resetting data as requested by %s."], sender))
			self:Reset()
		end
	end
end

function Violation:Listen(prefix, message, distribution, sender)
	if distribution ~= "RAID" or prefix ~= "DamageMeters" then return end

	local player, dmg = select(3, message:find("SYNC_12_ (%a+) (%d+)"))

	if tonumber(dmg) and type(player) == "string" and player ~= playerName and player == sender then
		damage[player] = tonumber(dmg)
		self:ScheduleEvent("DisplayUpdate", self.UpdateDisplay, 5, self)
	end
end

-----------------------------------------------------------------------
-- Combat log parsing
-----------------------------------------------------------------------
function Violation:OnDamageEvent(t)
	assert(t.sourceID == "player", "Not supposed to get combat events from people except ourselves.")
	if t.amount then
		damage[playerName] = damage[playerName] + t.amount
		damageDirty = damageDirty + t.amount
		self:UpdateDisplay()
	end
end

function Violation:OnHealEvent(t)
	assert(t.sourceID == "player", "Not supposed to get combat events from people except ourselves.")
	if t.amount then
		healing[playerName] = healing[playerName] + t.amount
		healingDirty = healingDirty + t.amount
		self:UpdateDisplay()
	end
end

